#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
import math
from theanolm.probfunctions import *

class TestProbFunctions(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_interpolate_linear(self):
        self.assertAlmostEqual(
            interpolate_linear(math.log(0.2), math.log(0.3), 0.25),
            math.log(0.25 * 0.2 + 0.75 * 0.3))
        self.assertAlmostEqual(
            interpolate_linear(float('-inf'), math.log(0.3), 0.01),
            math.log(0.3 * 0.99))
        self.assertEqual(
            interpolate_linear(float('-inf'), -10.0, 0.0),
            -10.0)
        self.assertAlmostEqual(
            interpolate_linear(math.log(0.3), float('-inf'), 0.99),
            math.log(0.3 * 0.99))
        self.assertEqual(
            interpolate_linear(-10.0, float('-inf'), 1.0),
            -10.0)
        self.assertAlmostEqual(
            interpolate_linear(-1001, -1002, 0.25),
            -1001.64263,  # ln(0.25 * exp(-1001) + 0.75 * exp(-1002))
            places=4)

    def test_interpolate_loglinear(self):
        self.assertEqual(
            interpolate_loglinear(-1001.0, -1002.0, 0.25, 0.75),
            -1001.75)
        self.assertEqual(
            interpolate_loglinear(float('-inf'), -1002.0, 0.25, 0.75),
            float('-inf'))
        self.assertEqual(
            interpolate_loglinear(float('-inf'), -1002.0, 0.0, 1.0),
            -1002.0)
        self.assertEqual(
            interpolate_loglinear(-1001.0, float('-inf'), 0.25, 0.75),
            float('-inf'))
        self.assertEqual(
            interpolate_loglinear(-1001.0, float('-inf'), 1.0, 0.0),
            -1001.0)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from abc import abstractmethod, ABCMeta
import logging
import numpy
import theano
import theano.tensor as tensor
from theanolm import Parameters
from theanolm.network.weightfunctions import random_matrix, matrix_from_value

class BasicLayer(object, metaclass=ABCMeta):
    """Superclass for Neural Network Layers
    """

    def __init__(self, layer_options, network, profile=False):
        """Saves some attributes that are common to all layers.

        :type layer_options: dict
        :param layer_options: dictionary of layer options

        :type network: Network
        :param network: the network object creating this layer

        :type profile: bool
        :param profile: if set to True, creates a Theano profile object
        """

        self.name = layer_options['name']
        self.input_layers = layer_options['input_layers']
        self.params = Parameters()
        self._devices = layer_options['devices']

        if 'size' in layer_options:
            self.output_size = int(layer_options['size'])
        else:
            self.output_size = \
                sum([x.output_size for x in self.input_layers])

        logging.debug("- %s name=%s inputs=[%s] size=%d, devices=[%s]",
            self.__class__.__name__,
            self.name,
            ', '.join([x.name for x in self.input_layers]),
            self.output_size,
            ', '.join([str(x) for x in self._devices]))

        self._network = network
        self._profile = profile

    @abstractmethod
    def create_structure(self):
        """Creates the symbolic graph of this layer.

        Sets self.output to a symbolic matrix that describes the output of this
        layer.
        """

        assert False

    def _param_path(self, param_name, device=None):
        """Returns the HDF5 path used to address a parameter.

        :type param_name: str
        :param param_name: name of a parameter within this layer

        :type device: str
        :param device: ``None`` for parameters that reside on the default device
                       only; otherwise returns the path used to address the part
                       of the parameter that resides on the given device

        :rtype: str
        :returns: full path of the parameter in a HDF5 file.
        """

        result = 'layers/' + self.name + '/' + param_name
        if not device is None:
            result += '/' + device
        return result

    def _get_param(self, param_name, device=None):
        """Returns a Theano tensor variable by parameter name.

        :type param_name: str
        :param param_name: name of a parameter within the layer

        :type device: str
        :param device: ``None`` for parameters that reside on the default device
                       only; otherwise returns the part of the parameter that
                       resides on the given device

        :rtype: TensorVariable
        :returns: the corresponding tensor variable
        """

        return self.params[self._param_path(param_name, device)]

    def _init_weight(self, param_name, shape, scale=None, count=1,
                     split_to_devices=False):
        """Generates a weight matrix from “standard normal” distribution.

        If ``shape`` contains two dimensions that match, generates an orthogonal
        matrix. In that case scale is ignored. Orthogonal weights are useful for
        two reasons:

        1. Multiplying by an orthogonal weight preserves the norm of the
           input vector, which should help avoid exploding and vanishing
           gradients.
        2. The row and column vectors are orthonormal to one another, which
           should help avoid two vectors learning to produce the same features.

        If ``count`` is specified, creates a concatenation of several similar
        submatrices (same shape but different content).

        If ``split_to_devices`` is set to ``True``, splits the weight to equal
        parts on the last dimension, and creates one parameter for each device.
        If also ``count`` is specified, each device will have an equal part of
        every submatrix.

        :type shape: list or tuple of ints
        :param shape: sizes of the weight dimensions; normally the first one is
                      the dimensionality of the input data and the second one is
                      the dimensionality of the output data

        :type scale: float
        :param scale: if other than ``None``, the matrix will be scaled by this
                      factor, unless an orthogonal matrix is created

        :type count: int
        :param count: concatenate this many weight matrices with the same shape

        :type split_to_devices: bool
        :param split_to_devices: if set to ``True``, creates on every device a
                                 parameter that contains one part of the weight
        """

        path = self._param_path(param_name)
        weight = random_matrix(shape, scale, count)
        if not split_to_devices:
            self.params.add(path, random_matrix(shape, scale, count))
        elif (len(self._devices) == 1) and (self._devices[0] == None):
            # This layer has not been assigned to a specific device.
            self.params.add(path, random_matrix(shape, scale, count))
        else:
            self._split_to_devices(path, weight, shape[-1])

    def _init_bias(self, param_name, shape, value=None, split_to_devices=False):
        """Initializes a bias vector with given value.

        If ``value`` is not given, initializes the vector with zero value. If
        ``value`` is a list, creates a concatenation of as many vectors as there
        are elements in the list.

        If ``split_to_devices`` is set to ``True``, splits the array to equal
        parts on the last dimension, and creates one parameter for each device.
        If ``value`` is a list, each device will have an equal part of every
        submatrix.

        :type param_name: str
        :param param_name: name for the parameter within the layer

        :type shape: int or tuple of ints
        :param shape: size of the vector, or a tuple of the sizes of each
                      dimension (in case ``value`` is a list, each part will
                      have this size)

        :type value: float, numpy.ndarray or list
        :param value: the value or array to initialize the elements to, or a
                      list of values or arrays to create a concatenation of
                      vectors

        :type split_to_devices: bool
        :param split_to_devices: if set to ``True``, creates on every device a
                                 parameter that contains one part of the array
        """

        path = self._param_path(param_name)
        bias = matrix_from_value(shape, value)
        if not split_to_devices:
            self.params.add(path, matrix_from_value(shape, value))
        elif (len(self._devices) == 1) and (self._devices[0] == None):
            # This layer has not been assigned to a specific device.
            self.params.add(path, matrix_from_value(shape, value))
        else:
            self._split_to_devices(path, bias, shape[-1])

    def _split_to_devices(self, path, value, part_size):
        """Splits a matrix to equal parts on the last dimension, and creates a
        parameter on each device.

        If the matrix consists of submatrices, each device will have an equal
        part of every submatrix, whose size is specified by ``part_size``.

        :type path: str
        :param path: base path for the parameters that will be prefixed by the
                     device string

        :type value: numpy.ndarray
        :param value: a matrix that will be split to give the initial value of
                      the parameters

        :type part_size: int
        :param part_size: size of the last dimension of ``value``, or if
                          ``value`` consists of multiple submatrices, size of
                          one submatrix
        """

        part_count = value.shape[-1] // part_size
        if part_count * part_size != value.shape[-1]:
            raise ValueError("Last dimension is not a multiple of part size.")

        split_sizes = self._size_per_device(part_size)
        split_start = 0
        for device, split_size in zip(self._devices, split_sizes):
            assert not device is None
            split_end = split_start + split_size
            ranges = []
            for part_index in range(part_count):
                part_start = part_index * part_size
                ranges.extend(range(part_start + split_start,
                                    part_start + split_end))
            split_start = split_end
            self.params.add(path + '/' + device, value[..., ranges], device)

    def _size_per_device(self, total_size):
        """Returns ``total_size`` divided for each device.

        :type total_size: int
        :param total_size: total size of a parameter

        :rtype: list of ints
        :returns: ``total_size`` divided into as many parts as there are devices
                  assigned to this layer
        """

        num_devices = len(self._devices)
        if num_devices < 1:
            raise RuntimeError("No devices assigned to this layer.")
        if total_size < num_devices:
            raise ValueError("Cannot split matrix of size {} to {} devices."
                             .format(total_size, num_devices))

        result = []
        quotient, remainder = divmod(total_size, num_devices)
        start_index = 0
        for i in range(1, num_devices + 1):
            end_index = i * quotient + min(i, remainder)
            result.append(end_index - start_index)
            start_index = end_index

        assert len(result) == num_devices
        assert sum(result) == total_size
        assert end_index == total_size

        return result

    def _tensor_preact(self, input_matrix, param_name):
        """Helper function that creates a pre-activation of ``input_matrix`` by
        multiplying it by a weight matrix and adding a bias.

        ``input_matrix`` and the result normally have the shape of a mini-batch:
        the first dimension is the time step and the second dimension is the
        sequence. The last dimension is always the data vector. The size of the
        input data vector should equal to the first dimension of the weight
        vector, and the second dimension of the weight vector defines the size
        of the output data vector.

        :type input_matrix: TensorVariable
        :param input_matrix: the preactivations will be computed by multiplying
                             the data vectors (the last dimension of this
                             matrix) by the weight matrix, and adding bias

        :type param_name: str
        :param param_name: name of a parameter group that contains a weight
                           matrix and a bias vector

        :rtype: TensorVariable
        :returns: a matrix tha has the same number of dimensions as
                  ``input_matrix``, but the data vectors (the last dimension of
                  this matrix) are the preactivations
        """

        weight = self.params[self._param_path(param_name) + '/W']
        bias = self.params[self._param_path(param_name) + '/b']
        return tensor.dot(input_matrix, weight) + bias



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy
from theanolm.parsing.linearbatchiterator import LinearBatchIterator

class ScoringBatchIterator(LinearBatchIterator):
    """Iterator for Reading Mini-Batches for Scoring Text

    Returns the actual words in addition to the word IDs. These are needed for
    subword combination. Only one file can be read.
    """

    def __init__(self, *args, **kwargs):
        """Constructs an iterator for reading mini-batches from given file or
        memory map.
        """

        super().__init__(*args, **kwargs)

    def _prepare_batch(self, sequences):
        """Transposes a list of sequences into a list of time steps. Then
        returns word ID and mask matrices in a format suitable to be input to
        the neural network, and words as a list of lists.

        The first returned matrix contains the word IDs, the second contains the
        word in plane text, and the third contains a mask that defines which
        elements are past the sequence end. Where the other values are valid,
        the mask matrix contains ones.

        The returned matrices have the same shape. The first dimensions is the
        time step, i.e. the index to a word in a sequence. The second dimension
        selects the sequence. In other words, the first row is the first word of
        each sequence and so on. However the plain text words are returned as a
        list of list of strs.

        :type sequences: list of lists
        :param sequences: list of sequences, each of which is a list of word
                          IDs

        :rtype: three ndarrays
        :returns: word ID, word, and mask structures
        """

        num_sequences = len(sequences)
        batch_length = numpy.max([len(s) for s in sequences])

        unk_id = self.vocabulary.word_to_id['<unk>']
        shape = (batch_length, num_sequences)
        word_ids = numpy.ones(shape, numpy.int64) * unk_id
        words = []
        mask = numpy.zeros(shape, numpy.int8)

        for i, sequence in enumerate(sequences):
            length = len(sequence)

            sequence_word_ids = numpy.ones(length, numpy.int64) * unk_id
            sequence_words = []
            for index, (word, file_id) in enumerate(sequence):
                if word in self.vocabulary.word_to_id:
                    sequence_word_ids[index] = self.vocabulary.word_to_id[word]
                sequence_words.append(word)

            word_ids[:length, i] = sequence_word_ids
            words.append(sequence_words)
            mask[:length, i] = 1

        return word_ids, words, mask



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import subprocess
import numpy
import theano
from theanolm import Network
from theanolm.parsing import ScoringBatchIterator
from theanolm.scoring import TextScorer
from theanolm.filetypes import TextFileType

def add_arguments(parser):
    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to score text')
    argument_group.add_argument(
        'input_file', metavar='TEXT-FILE', type=TextFileType('r'),
        help='text file containing text to be scored (UTF-8, one sentence per '
             'line, assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the statistics (default stdout, will be '
             'compressed if the name ends in ".gz")')

    argument_group = parser.add_argument_group("scoring")
    argument_group.add_argument(
        '--output', metavar='DETAIL', type=str, default='perplexity',
        help='what to output, one of "perplexity", "utterance-scores", '
             '"word-scores" (default "perplexity")')
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help='convert output log probabilities to base B (default is the '
             'natural logarithm)')
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help='if LOGPROB is zero, do not include <unk> tokens in perplexity '
             'computation; otherwise use constant LOGPROB as <unk> token score '
             '(default is to use the network to predict <unk> probability)')
    argument_group.add_argument(
        '--subwords', metavar='MARKING', type=str, default=None,
        help='the subword vocabulary uses MARKING to indicate how words are '
             'formed from subwords; one of "word-boundary" (<w> token '
             'separates words), "prefix-affix" (subwords that can be '
             'concatenated are prefixed or affixed with +, e.g. "cat+ +s")')

def score(args):
    network = Network.from_file(args.model_path)

    print("Building text scorer.")
    sys.stdout.flush()
    if args.unk_penalty is None:
        ignore_unk = False
        unk_penalty = None
    elif args.unk_penalty == 0:
        ignore_unk = True
        unk_penalty = None
    else:
        ignore_unk = False
        unk_penalty = args.unk_penalty
    scorer = TextScorer(network, ignore_unk, unk_penalty)

    print("Scoring text.")
    if args.output == 'perplexity':
        _score_text(args.input_file, network.vocabulary, scorer,
                    args.output_file, args.log_base, args.subwords, False)
    elif args.output == 'word-scores':
        _score_text(args.input_file, network.vocabulary, scorer,
                    args.output_file, args.log_base, args.subwords, True)
    elif args.output == 'utterance-scores':
        _score_utterances(args.input_file, network.vocabulary, scorer,
                          args.output_file, args.log_base)
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)

def _score_text(input_file, vocabulary, scorer, output_file,
                log_base=None, subword_marking=None, word_level=False):
    """Reads text from ``input_file``, computes perplexity using
    ``scorer``, and writes to ``output_file``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base

    :type subword_marking: str
    :param subword_marking: if other than None, vocabulary is subwords;
        "word-boundary" indicates <w> token separates words, "prefix-affix"
        indicates subwords are prefixed/affixed with +

    :type word_level: bool
    :param word_level: if set to True, also writes word-level statistics
    """

    validation_iter = \
        ScoringBatchIterator(input_file,
                             vocabulary,
                             batch_size=16,
                             max_sequence_length=None)
    log_scale = 1.0 if log_base is None else numpy.log(log_base)

    total_logprob = 0.0
    num_sentences = 0
    num_tokens = 0
    num_words = 0
    num_unks = 0
    num_probs = 0
    for word_ids, words, mask in validation_iter:
        class_ids, membership_probs = vocabulary.get_class_memberships(word_ids)
        logprobs = scorer.score_batch(word_ids, class_ids, membership_probs,
                                      mask)
        for seq_index, seq_logprobs in enumerate(logprobs):
            seq_word_ids = word_ids[:, seq_index]
            seq_mask = mask[:, seq_index]
            seq_word_ids = seq_word_ids[seq_mask == 1]
            seq_words = words[seq_index]
            merged_words, merged_logprobs = _merge_subwords(seq_words,
                                                            seq_logprobs,
                                                            subword_marking)

            # total logprob of this sequence
            seq_logprob = sum(filter(None, merged_logprobs))
            # total logprob of all sequences
            total_logprob += seq_logprob
            # number of tokens, which may be subwords, including <unk>'s
            num_tokens += len(seq_word_ids)
            # number of words, including <unk>'s
            num_words += len(merged_words)
            # number of word probabilities computed (may not include <unk>'s)
            num_probs += sum(not x is None for x in merged_logprobs)
            # number of <unk>'s (just for reporting)
            num_unks += sum(x is None for x in merged_logprobs)
            # number of sequences
            num_sentences += 1

            if word_level:
                output_file.write("# Sentence {0}\n".format(num_sentences))
                _write_word_scores(merged_words, merged_logprobs, output_file,
                                   log_scale)
                output_file.write("Sentence perplexity: {0}\n\n".format(
                    numpy.exp(-seq_logprob / len(seq_logprobs))))

    output_file.write("Number of sentences: {0}\n".format(num_sentences))
    output_file.write("Number of words: {0}\n".format(num_words))
    output_file.write("Number of tokens: {0}\n".format(num_tokens))
    output_file.write("Number of predicted probabilities: {0}\n".format(num_probs))
    output_file.write("Number of excluded (OOV) words: {0}\n".format(num_unks))
    if num_words > 0:
        cross_entropy = -total_logprob / num_probs
        perplexity = numpy.exp(cross_entropy)
        output_file.write("Cross entropy (base e): {0}\n".format(cross_entropy))
        if not log_base is None:
            cross_entropy /= log_scale
            output_file.write("Cross entropy (base {1}): {0}\n".format(
                cross_entropy, log_base))
        output_file.write("Perplexity: {0}\n".format(perplexity))

def _merge_subwords(subwords, subword_logprobs, marking):
    """Creates a word list from a subword list.

    :type subwords: list of strs
    :param subwords: list of vocabulary words, which may be subwords

    :type subword_logprobs: list of floats
    :param subword_logprobs: list of log probabilities for each word/subword
                             starting from the second one, containing ``None``
                             in place of any ignored <unk>'s

    :type marking: str
    :param marking: ``None`` for word vocabulary, otherwise the type of subword
                    marking used: "word-boundary" if a word boundary token (<w>)
                    is used and "prefix-affix" if subwords are prefixed/affixed
                    with + when they can be concatenated

    :rtype: tuple of (list of strs, list of floats)
    :returns: the first item is a list of words; the second item is a list of
              log probabilities for each word starting from the second one,
              containing ``None`` in place of any ignored <unk>'s
    """

    if len(subword_logprobs) != len(subwords) - 1:
        raise ValueError("Number of logprobs should be exactly one less than "
                         "the number of words.")

    if marking is None:
        # Vocabulary is already words.
        return subwords, subword_logprobs

    words = [subwords[0]]
    logprobs = []
    current_word = ''
    current_logprob = 0.0

    if marking == 'word-boundary':
        # Words are separated by <w>. Merge subwords and logprobs between <w>
        # tokens. If any part of a word is <unk>, the whole word is <unk>.
        for subword, logprob in zip(subwords[1:], subword_logprobs):
            if (current_logprob is None) or (logprob is None):
                current_logprob = None
            else:
                current_logprob += logprob
            if subword == '<w>':
                if current_word != '':
                    words.append(current_word)
                    logprobs.append(current_logprob)
                    current_word = ''
                current_logprob = 0.0
            elif (current_word == '<unk>') or (subword == '<unk>'):
                current_word = '<unk>'
            else:
                current_word += subword
    elif marking == 'prefix-affix':
        # Merge subword to the current word if the current word ends in + and
        # the subword starts with +.
        for subword, logprob in zip(subwords[1:], subword_logprobs):
            if current_word.endswith('+') and subword.startswith('+'):
                current_word = current_word[:-1] + subword[1:]
                if (current_logprob is None) or (logprob is None):
                    current_logprob = None
                else:
                    current_logprob += logprob
            else:
                if current_word != '':
                    words.append(current_word)
                    logprobs.append(current_logprob)
                current_word = subword
                current_logprob = logprob
    else:
        raise ValueError("Invalid subword marking type: " + marking)

    if current_word != '':
        words.append(current_word)
        logprobs.append(current_logprob)
    return words, logprobs

def _write_word_scores(words, logprobs, output_file, log_scale):
    """Writes word-level scores to an output file.

    :type words: list of strs
    :param words: sequence of words

    :type logprobs: list of floats
    :param logprobs: logprob of each word starting from the second word

    :type output_file: file object
    :param output_file: a file where to write the output

    :type log_scale: float
    :param log_scale: divide logprobs by this amount to convert to correct base
    """

    if len(logprobs) != len(words) - 1:
        raise ValueError("Number of logprobs should be exactly one less than "
                         "the number of words.")

    logprobs = [None if x is None else x / log_scale for x in logprobs]
    for index, logprob in enumerate(logprobs):
        if index - 2 > 0:
            history = ['...']
            history.extend(words[index - 2:index + 1])
        else:
            history = words[:index + 1]
        history = ' '.join(history)
        predicted = words[index + 1]

        if logprob is None:
            output_file.write("p({0} | {1}) is not predicted\n".format(
                predicted, history))
        else:
            output_file.write("log(p({0} | {1})) = {2}\n".format(
                predicted, history, logprob))

def _score_utterances(input_file, vocabulary, scorer, output_file,
                      log_base=None):
    """Reads utterances from ``input_file``, computes LM scores using
    ``scorer``, and writes one score per line to ``output_file``.

    Start-of-sentence and end-of-sentece tags (``<s>`` and ``</s>``) will be
    inserted at the beginning and the end of each utterance, if they're missing.
    Empty lines will be ignored, instead of interpreting them as the empty
    sentence ``<s> </s>``.

    :type input_file: file object
    :param input_file: a file that contains the input sentences in SRILM n-best
                       format

    :type vocabulary: Vocabulary
    :param vocabulary: vocabulary that provides mapping between words and word
                       IDs

    :type scorer: TextScorer
    :param scorer: a text scorer for rescoring the input sentences

    :type output_file: file object
    :param output_file: a file where to write the output n-best list in SRILM
                        format

    :type log_base: int
    :param log_base: if set to other than None, convert log probabilities to
                     this base
    """

    log_scale = 1.0 if log_base is None else numpy.log(log_base)

    for line_num, line in enumerate(input_file):
        lm_score = scorer.score_line(line, vocabulary)
        if lm_score is None:
            continue
        lm_score /= log_scale
        output_file.write(str(lm_score) + '\n')
        if (line_num + 1) % 1000 == 0:
            print("{0} sentences scored.".format(line_num + 1))
        sys.stdout.flush()

    if scorer.num_words == 0:
        print("The input file contains no words.")
    else:
        print("{0} words processed, including start-of-sentence and "
              "end-of-sentence tags, and {1} ({2:.1f} %) out-of-vocabulary "
              "words".format(scorer.num_words,
                             scorer.num_unks,
                             scorer.num_unks / scorer.num_words))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import logging
import subprocess
import numpy
import theano
from theanolm import Network
from theanolm.scoring import LatticeDecoder, SLFLattice
from theanolm.filetypes import TextFileType

def add_arguments(parser):
    argument_group = parser.add_argument_group("files")
    argument_group.add_argument(
        'model_path', metavar='MODEL-FILE', type=str,
        help='the model file that will be used to decode the lattice')
    argument_group.add_argument(
        '--lattices', metavar='FILE', type=str, nargs='*', default=[],
        help='word lattices to be decoded (SLF, assumed to be compressed if '
             'the name ends in ".gz")')
    argument_group.add_argument(
        '--lattice-list', metavar='FILE', type=TextFileType('r'),
        help='text file containing a list of word lattices to be decoded (one '
             'path to an SLF file per line, the list and the SLF files are '
             'assumed to be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--output-file', metavar='FILE', type=TextFileType('w'), default='-',
        help='where to write the best paths through the lattices (default '
             'stdout, will be compressed if the name ends in ".gz")')
    argument_group.add_argument(
        '--num-jobs', metavar='J', type=int, default=1,
        help='divide the set of lattice files into J distinct batches, and '
             'process only batch I')
    argument_group.add_argument(
        '--job', metavar='I', type=int, default=0,
        help='the index of the batch that this job should process, between 0 '
             'and J-1')

    argument_group = parser.add_argument_group("decoding")
    argument_group.add_argument(
        '--output', metavar='FORMAT', type=str, default='ref',
        help='format of the output, one of "ref" (default, utterance ID '
             'followed by words), "trn" (words followed by utterance ID in '
             'parentheses), "full" (utterance ID, acoustic score, language '
             'score, and number of words, followed by words)')
    argument_group.add_argument(
        '--n-best', metavar='N', type=int, default=1,
        help='print N best paths of each lattice (default 1)')
    argument_group.add_argument(
        '--nnlm-weight', metavar='LAMBDA', type=float, default=1.0,
        help="language model probabilities given by the model read from "
             "MODEL-FILE will be weighted by LAMBDA, when interpolating with "
             "the language model probabilities in the lattice (default is 1.0, "
             "meaning that the LM probabilities in the lattice will be "
             "ignored)")
    argument_group.add_argument(
        '--lm-scale', metavar='LMSCALE', type=float, default=None,
        help="scale language model log probabilities by LMSCALE when computing "
             "the total probability of a path (default is to use the LM scale "
             "specified in the lattice file, or 1.0 if not specified)")
    argument_group.add_argument(
        '--wi-penalty', metavar='WIP', type=float, default=None,
        help="penalize word insertion by adding WIP to the total log "
             "probability as many times as there are words in the path "
             "(without scaling WIP by LMSCALE)")
    argument_group.add_argument(
        '--log-base', metavar='B', type=int, default=None,
        help="convert output log probabilities to base B and WIP from base B "
             "(default is natural logarithm; this does not affect reading "
             "lattices, since they specify their internal log base)")
    argument_group.add_argument(
        '--unk-penalty', metavar='LOGPROB', type=float, default=None,
        help="if LOGPROB is zero, do not include <unk> tokens in perplexity "
             "computation; otherwise use constant LOGPROB as <unk> token score "
             "(default is to use the network to predict <unk> probability)")
    argument_group.add_argument(
        '--linear-interpolation', action="store_true",
        help="use linear interpolation of language model probabilities, "
             "instead of (pseudo) log-linear")

    argument_group = parser.add_argument_group("pruning")
    argument_group.add_argument(
        '--max-tokens-per-node', metavar='T', type=int, default=None,
        help="keep only at most T tokens at each node when decoding a lattice "
             "(default is no limit)")
    argument_group.add_argument(
        '--beam', metavar='B', type=float, default=None,
        help="prune tokens whose log probability is at least B smaller than "
             "the log probability of the best token at any given time (default "
             "is no beam pruning)")
    argument_group.add_argument(
        '--recombination-order', metavar='O', type=int, default=None,
        help="keep only the best token, when at least O previous words are "
             "identical (default is to recombine tokens only if the entire "
             "word history matches)")

    argument_group = parser.add_argument_group("logging and debugging")
    argument_group.add_argument(
        '--log-file', metavar='FILE', type=str, default='-',
        help='path where to write log file (default is standard output)')
    argument_group.add_argument(
        '--log-level', metavar='LEVEL', type=str, default='info',
        help='minimum level of events to log, one of "debug", "info", "warn" '
             '(default "info")')
    argument_group.add_argument(
        '--debug', action="store_true",
        help='enables debugging Theano errors')
    argument_group.add_argument(
        '--profile', action="store_true",
        help='enables profiling Theano functions')

def decode(args):
    log_file = args.log_file
    log_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(log_level, int):
        print("Invalid logging level requested:", args.log_level)
        sys.exit(1)
    log_format = '%(asctime)s %(funcName)s: %(message)s'
    if args.log_file == '-':
        logging.basicConfig(stream=sys.stdout, format=log_format, level=log_level)
    else:
        logging.basicConfig(filename=log_file, format=log_format, level=log_level)

    if args.debug:
        theano.config.compute_test_value = 'warn'
    else:
        theano.config.compute_test_value = 'off'
    theano.config.profile = args.profile
    theano.config.profile_memory = args.profile

    network = Network.from_file(args.model_path)

    log_scale = 1.0 if args.log_base is None else numpy.log(args.log_base)

    if args.wi_penalty is None:
        wi_penalty = None
    else:
        wi_penalty = args.wi_penalty * log_scale
    if args.unk_penalty is None:
        ignore_unk = False
        unk_penalty = None
    elif args.unk_penalty == 0:
        ignore_unk = True
        unk_penalty = None
    else:
        ignore_unk = False
        unk_penalty = args.unk_penalty
    decoding_options = {
        'nnlm_weight': args.nnlm_weight,
        'lm_scale': args.lm_scale,
        'wi_penalty': wi_penalty,
        'ignore_unk': ignore_unk,
        'unk_penalty': unk_penalty,
        'linear_interpolation': args.linear_interpolation,
        'max_tokens_per_node': args.max_tokens_per_node,
        'beam': args.beam,
        'recombination_order': args.recombination_order
    }
    logging.debug("DECODING OPTIONS")
    for option_name, option_value in decoding_options.items():
        logging.debug("%s: %s", option_name, str(option_value))

    print("Building word lattice decoder.")
    sys.stdout.flush()
    decoder = LatticeDecoder(network, decoding_options)

    # Combine paths from command line and lattice list.
    lattices = args.lattices
    if not args.lattice_list is None:
        lattices.extend(args.lattice_list.readlines())
    lattices = [path.strip() for path in lattices]
    # Ignore empty lines in the lattice list.
    lattices = list(filter(None, lattices))
    # Pick every Ith lattice, if --num-jobs is specified and > 1.
    if args.num_jobs < 1:
        print("Invalid number of jobs specified:", args.num_jobs)
        sys.exit(1)
    if (args.job < 0) or (args.job > args.num_jobs - 1):
        print("Invalid job specified:", args.job)
        sys.exit(1)
    lattices = lattices[args.job::args.num_jobs]

    file_type = TextFileType('r')
    for index, path in enumerate(lattices):
        logging.info("Reading word lattice: %s", path)
        lattice_file = file_type(path)
        lattice = SLFLattice(lattice_file)

        if not lattice.utterance_id is None:
            utterance_id = lattice.utterance_id
        else:
            utterance_id = os.path.basename(lattice_file.name)
        logging.info("Utterance `%s' -- %d/%d of job %d",
                     utterance_id,
                     index + 1,
                     len(lattices),
                     args.job)
        tokens = decoder.decode(lattice)

        for index in range(min(args.n_best, len(tokens))):
            line = format_token(tokens[index],
                                utterance_id,
                                network.vocabulary,
                                log_scale,
                                args.output)
            args.output_file.write(line + "\n")

def format_token(token, utterance_id, vocabulary, log_scale, format):
    """Formats an output line from a token and an utterance ID.

    Reads word IDs from the history list of ``token`` and converts them to words
    using ``vocabulary``. The history may contain also OOV words as text, so any
    ``str`` will be printed literally.

    :type token: Token
    :param token: a token whose history will be formatted

    :type utterance_id: str
    :param utterance_id: utterance ID for full output

    :type vocabulary: Vocabulary
    :param vocabulary: mapping from word IDs to words

    :type log_scale: float
    :param log_scale: divide log probabilities by this number to convert the log
                      base

    :type format: str
    :param format: which format to write, one of "ref" (utterance ID, words),
                   "trn" (words, utterance ID in parentheses), "full" (utterance
                   ID, acoustic and LM scores, number of words, words)

    :rtype: str
    :returns: the formatted output line
    """

    words = token.history_words(vocabulary)
    if format == 'ref':
        return "{} {}".format(utterance_id, ' '.join(words))
    elif format == 'trn':
        return "{} ({})".format(' '.join(words), utterance_id)
    elif format == 'full':
        return "{} {} {} {} {}".format(
            utterance_id,
            token.ac_logprob / log_scale,
            token.lm_logprob / log_scale,
            len(words),
            ' '.join(words))
    else:
        print("Invalid output format requested:", args.output)
        sys.exit(1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from theanolm.commands.score import _merge_subwords
from numpy.testing import assert_almost_equal

class TestScore(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_merge_subwords(self):
        # word vocabulary
        subwords = ['<s>', 'aaa', 'bbb', 'ccc', 'ddd', '</s>']
        subword_logprobs = [0.0, 0.1, 0.2, 0.3, 0.4]
        words, word_logprobs = _merge_subwords(subwords, subword_logprobs, None)
        self.assertSequenceEqual(words, subwords)
        assert_almost_equal(word_logprobs, subword_logprobs)

        # subword vocabulary with word boundary token, <unk> predicted
        subwords = ['<s>', '<w>', 'aaa', '<w>', 'bbb', '<unk>', '<w>', 'ccc', 'ddd', '<w>', '</s>']
        subword_logprobs = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        words, word_logprobs = _merge_subwords(subwords, subword_logprobs, "word-boundary")
        self.assertSequenceEqual(words, ['<s>', 'aaa', '<unk>', 'cccddd', '</s>'])
        assert_almost_equal(word_logprobs, [0.2 + 0.3, 0.4 + 0.5 + 0.6, 0.7 + 0.8 + 0.9, 1.0])

        # subword vocabulary with word boundary token, <unk> not predicted
        subword_logprobs = [0.1, 0.2, 0.3, 0.4, None, 0.6, 0.7, 0.8, 0.9, 1.0]
        words, word_logprobs = _merge_subwords(subwords, subword_logprobs, "word-boundary")
        self.assertSequenceEqual(words, ['<s>', 'aaa', '<unk>', 'cccddd', '</s>'])
        self.assertAlmostEqual(word_logprobs[0], 0.2 + 0.3)
        self.assertIsNone(word_logprobs[1])
        self.assertAlmostEqual(word_logprobs[2], 0.7 + 0.8 + 0.9)
        self.assertAlmostEqual(word_logprobs[3], 1.0)

        # subword vocabulary with prefix/affix markings, <unk> predicted
        subwords = ['<s>', 'aaa', 'bbb+', '+ccc', '<unk>', '</s>']
        subword_logprobs = [0.1, 0.2, 0.3, 0.4, 0.5]
        words, word_logprobs = _merge_subwords(subwords, subword_logprobs, "prefix-affix")
        self.assertSequenceEqual(words, ['<s>', 'aaa', 'bbbccc', '<unk>', '</s>'])
        assert_almost_equal(word_logprobs, [0.1, 0.2 + 0.3, 0.4, 0.5])

        # subword vocabulary with prefix/affix markings, <unk> not predicted
        subwords = ['<s>', 'aaa', 'bbb+', '+ccc', '<unk>', '</s>']
        subword_logprobs = [0.1, 0.2, 0.3, None, 0.5]
        words, word_logprobs = _merge_subwords(subwords, subword_logprobs, "prefix-affix")
        self.assertSequenceEqual(words, ['<s>', 'aaa', 'bbbccc', '<unk>', '</s>'])
        self.assertAlmostEqual(word_logprobs[0], 0.1)
        self.assertAlmostEqual(word_logprobs[1], 0.2 + 0.3)
        self.assertIsNone(word_logprobs[2])
        self.assertAlmostEqual(word_logprobs[3], 0.5)

if __name__ == '__main__':
    unittest.main()
